use Module::Build;

my $build = Module::Build->new(
    'module_name'    => 'Debug::Easy',
    'license'        => 'perl',
    'configure_requires' => {
        'Module::Build' => '>= 0.30'
    },
    'requires'       => {
        'perl'                      => '5.6.1',
        'Log::Fast'                 => '>= 1.00',
        'Time::HiRes'               => '>= 1.00',
        'Term::ANSIColor'           => '>= 4.00',
        'Module::Load::Conditional' => '>= 0.50',
    },
#    'recommends'     => {
#        'Data::Printer'        => '>= 0.30',
#        'Data::Dumper::Simple' => '>= 0.10',
#    },
    'build_requires' => {
        'Test::More' => '> 0'
    },
    'dist'           => {
        'COMPRESS' => 'gzip -9f',
        'SUFFIX'   => 'gz'
    },
    'clean'          => {
        'FILES' => 'Debug-Easy-*'
    }
);

$build->create_build_script();

